/*

Based on comp1.c from Michael Schindler. Original copyright follows.

This program parses position, cigar string and sequence from a file
and encodes the sequence as a delta against a known reference.



  (c) Michael Schindler
  1999
  http://www.compressconsult.com/
  michael@compressconsult.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA.

  comp1 is an example compressor trying to compress files with a simple
  order 1 model. The files can be decompressed by decomp1.

  Note that I do not think that an order 1 model as here is good;
  For better compression see for example my freeware szip.
  http://www.compressconsult.com/szip/
  or ask me as consultant what compression method fits your data best.
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "port.h"
#include "qsmodel.h"
#include "rangecod.h"

/* Define REF if you wish it to use the reference for anything more
 * than GC composition stats
 */
//#define REF

static unsigned char *load(int *lenp, char *fn) {
    unsigned char *data = NULL;
    int dsize = 0;
    int dcurr = 0, len;
    int fd = 0;

    if (fn) {
	if (-1 == (fd = open(fn, O_RDONLY, 0))) {
	    perror(fn);
	    return NULL;
	}
    }

    do {
	if (dsize - dcurr < 8192) {
	    dsize = dsize ? dsize * 2 : 8192;
	    if (NULL == (data = realloc(data, dsize))) {
		if (fd)
		    close(fd);
		return NULL;
	    }
	}

	len = read(fd, data + dcurr, 8192);
	if (len > 0)
	    dcurr += len;
    } while (len > 0);

    if (len == -1) {
	perror("read");
	if (fd)
	    close(fd);
	return NULL;
    }

    if (fd)
	close(fd);

    *lenp = dcurr;
    return data;
}

#define MAX_LINE_LEN 65536
int main( int argc, char *argv[] )
{   int ch, syfreq, ltfreq, lastchar=0;
    rangecoder rc;
    char *ref = NULL;
    int ref_len, i;
    qsmodel *ref_model, ins, soft2[25], ins2[25], ins_model[100], mat;
    int lookup[256];
    char line[MAX_LINE_LEN], seq[MAX_LINE_LEN];
    int nmis = 0;
    int nins = 0;
    int nsoft = 0;
    int nbases = 0;
    int last_ch = 0;
    FILE *fp;
    int ref_freq[5];
    char *ref_init;
    int count = 0;

#ifdef REF
    if (argc != 3) {
	fprintf(stderr, "Usage: sam_comp_ref pos_cigar.txt reference.raw < seq > seq.comp\n");
	exit(1);
    }
#else
    if (argc != 3) {
	fprintf(stderr, "Usage: sam_comp pos_cigar.txt ref_length < seq > seq.comp\n");
	exit(1);
    }
#endif

    for (i = 0; i < 256; i++)
	lookup[i] = 4;
    lookup['A'] = 0;     lookup['a'] = 0; 
    lookup['C'] = 1;     lookup['c'] = 1; 
    lookup['G'] = 2;     lookup['g'] = 2; 
    lookup['T'] = 3;     lookup['t'] = 3; 

    /* Load the data sets */
#ifdef REF
    if (NULL == (ref = load(&ref_len, argv[2])))
	exit(1);
#else
    ref_len = atoi(argv[2]);
#endif
    if (NULL == (fp = fopen(argv[1], "r"))) {
	perror(argv[1]);
	exit(1);
    }

#ifndef REF
    /* Gather stats on reference, these would need transmitting with seq */
    for (i = 0; i < 5; i++) 
	ref_freq[i] = 1;

    /*
     * Gather on-the-fly instead
     */
    /*
    for (i = 0; i < ref_len; i++) {
	ref_freq[lookup[ref[i]]]++;
    }

    for (i = 0; i < 5; i++)
	fprintf(stderr, "freq[%d] = %d\n", i, ref_freq[i]);

    */
    ref_init  = (char *)calloc(ref_len, 1);
#endif

    /* Initialise our models */
#define NB 9
    ref_model = (qsmodel *)calloc(ref_len, sizeof(qsmodel));
    for (i = 0; i < ref_len; i++) {
	int init[5];
	int tot, j;

#ifdef REF
	if (ref[i] == 'N') {
	    init[0] = init[1] = init[2] = init[3] = 10;
	    init[4] = (1<<NB) - 4*init[0];
	} else {
	    init[0] = init[1] = init[2] = init[3] = 7; init[4] = 1;
	    init[lookup[ref[i]]] = (1<<NB)-3*init[0]-1;
	}
	initqsmodel(&ref_model[i], 5, NB, 1, init, 1);
#endif

#ifndef REF
#if 0
	/* Or with no reference, flat priors */
	initqsmodel(&ref_model[i], 5, NB, 1, 0, 1);
#else
	/* Or with no reference, using prior knowledge of GC content */
	tot = 0;
	for (j = 0; j < 5; j++) {
	    init[j] = (1<<NB) * ref_freq[j] / ref_len;
	    tot += init[j];
	}
	init[4] = (1<<NB)-tot;
	initqsmodel(&ref_model[i], 5, NB, 1, init, 1);
#endif
#endif
    }

    initqsmodel(&ins, 5, NB, 50, NULL, 1);

    /* Order 2 predictors for soft-clip and insertion seqs. UNUSED */
    for (i = 0; i < 25; i++)
	initqsmodel(&soft2[i], 5, NB, 50, NULL, 1);
    for (i = 0; i < 25; i++)
	initqsmodel(&ins2[i], 5, NB, 50, NULL, 1);

    /* Loop around, parsing line by line */
    start_encoding(&rc,0,0);

    last_ch = 0;
    while (fgets(line, MAX_LINE_LEN, fp)) {
	char *cp = line;
	char *cigar;
	int pos, seq_len;
	int cigar_op;
	int cigar_len;
	int start;

	pos = strtol(cp, &cp, 10)-1;
	cp++; /* skip tab */
	
	cigar = cp;
	while (*cp && *cp != '\n')
	    cp++;
	*cp++ = 0;

	if (NULL == (fgets(seq, MAX_LINE_LEN, stdin))) {
	    fprintf(stderr, "EOF on input seq file\n");
	    exit(1);
	}
	cp = seq;
	while (*cp && *cp != '\n')
	    cp++;
	*cp = 0;

	//printf("%d:%s:%s:\n", pos, cigar, seq);

	cigar_len = strtol(cigar, &cigar, 10);
	cigar_op = *cigar++;

	/* do the encoding */
	seq_len = strlen(seq);
	nbases += seq_len;

	/*
	if (memcmp(seq, &ref[pos], seq_len) == 0) {
	    qsgetfreq(&mat,1,&syfreq,&ltfreq);
	    encode_shift(&rc,syfreq,ltfreq,NB);
	    qsupdate(&mat,1);
	} else {
	    qsgetfreq(&mat,0,&syfreq,&ltfreq);
	    encode_shift(&rc,syfreq,ltfreq,NB);
	    qsupdate(&mat,0);
	*/
	//start = 1;
	for (i = 0; i < seq_len; i++) {
	    char ch = lookup[seq[i]];
	    qsmodel *qsm;

#ifndef REF
	    /* Predict of frequencies for next (unknown) base */
	    if (!ref_init[pos]) {
		int init[5], j, tot, sum;

		/*
		for (sum = j = 0; j < 5; j++) {
		    sum += ref_freq1[last_ch*5 + j];
		}
		for (tot = j = 0; j < 4; j++) {
		    init[j] = (1<<NB) * ref_freq1[last_ch*5 + j] / sum;
		    if (init[j] == 0)
			init[j] = 1;
		    tot += init[j];
		}
		init[4] = (1<<NB)-tot;
		*/

		for (sum = j = 0; j < 5; j++) {
		    sum += ref_freq[j];
		}
		for (tot = j = 0; j < 4; j++) {
		    init[j] = (1<<NB) * ref_freq[j] / sum;
		    if (init[j] == 0)
			init[j] = 1;
		    tot += init[j];
		}
		init[4] = (1<<NB)-tot;

		initqsmodel(&ref_model[pos], 5, NB, 1, init, 1);
		//fprintf(stderr, "init %d = {%d,%d,%d,%d,%d}\n", pos,
		//	init[0], init[1], init[2], init[3], init[4]);
		ref_init[pos] = 1;
	    }
#endif

	    switch(cigar_op) {
	    case 'S':
		nsoft++;
		//printf("cigar_len=%d\n", cigar_len);
		/*
		if (start) {
		    qsm = &soft2[lookup[ref[pos-cigar_len]]];
		} else {
		    qsm = &soft2[lookup[ref[pos+cigar_len+1]]];
		}
		*/
		qsm = &ins;
		
		/* order-2 prediction of next symbol */
		//qsm = &soft2[last_ch];
		break;

	    case 'I':
		start = 0;
		nins++;
		//printf("S/I %c\n", seq[i]);
		qsm = &ins;

		//qsm = &ins2[last_ch];
		break;

	    case 'D':
		start = 0;
		pos++;
		i--;
		goto next;

	    case 'M':
		start = 0;
		//printf("seq[%d]=%c ref[%d]=%c %c\n", i, seq[i], pos, ref[pos],
		//       "! "[seq[i] == ref[pos]]);
#ifdef REF
		if (seq[i] != ref[pos])
		    nmis++;
#endif
		qsm = &ref_model[pos];
		pos++;
		break;

	    default:
		exit(1);
	    }

	    //printf("ch=%d/%c\n", ch,"ACGTN"[ch]);

	    qsgetfreq(qsm,ch,&syfreq,&ltfreq);
	    encode_shift(&rc,syfreq,ltfreq,NB);
	    qsupdate(qsm,ch);

#ifndef REF
	    ref_freq[ch]++;
	    if (count++ == 5000) {
		int j;
		count = 0;
		for (j = 0; j < 5; j++) {
		    //fprintf(stderr, "%d : %d -> %d\n",
		    //	    j, ref_freq[j], ref_freq[j] / 2 + 1);
		    ref_freq[j] = ref_freq[j] / 2 + 1;
		}
	    }

	    /*
	    ref_freq1[last_ch*5 + ch]++;
	    if (count++ == 500000) {
		int j;
		count = 0;
		for (j = 0; j < 25; j++) {
		    //fprintf(stderr, "%d : %d -> %d\n",
		    //	    j, ref_freq1[j], ref_freq1[j] / 2 + 1);
		    ref_freq1[j] = ref_freq1[j]/2 + 1;
		}
	    }
	    */
#endif

	    //last_ch = (last_ch * 5 + ch) % 25;
	    last_ch = ch;

	next:
	    if (--cigar_len == 0) {
		cigar_len = strtol(cigar, &cigar, 10);
		cigar_op = *cigar++;
	    }
	}
	/* } */
    }

    fprintf(stderr, "N.bases    %d\n", nbases);
    fprintf(stderr, "mismatches %d\n", nmis);
    fprintf(stderr, "insertions %d\n", nins);
    fprintf(stderr, "soft-clips %d\n", nsoft);
    fprintf(stderr, "=> matched %d\n", nbases -nmis-nins-nsoft);

    /* write 256 as end-of-file */
    //qsgetfreq(qsm+lastchar,256,&syfreq,&ltfreq);
    //encode_shift(&rc,syfreq,ltfreq,NB);

    done_encoding(&rc);
    //for (ch=0; ch<256; ch++)
    //    deleteqsmodel(qsm+ch);

    return 0;
}

/*
 * sam_comp3 -c zlib -b 100000: 359571
 * sam_comp3 -c lzma -b 100000: 274156
 * this                         231737
 *
 * sam_comp3 -r -c zlib -b 100000: 213619
 * sam_comp3 -r -c lzma -b 100000: 180200
 * this                            150273
 */
